package com.holub.ui;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;

import com.holub.ui.AncestorAdapter;

/**
 *	The Bag is a Collection that supports a notion of a user interface.
 *	The Bag can produce a "visual proxy" that will take the form
 *	of a JList, JComboBox, or JButton (which displays a dialog when
 *	pushed), depending on the amount of screen real estate that's
 *	available to it. The proxy decides how to display itself without
 *	any outside intervention. Moreover, the "visual proxy" automatically
 *	reflects model-level changes such as adding or removing items from the list.
 *	For example, when you add or remove an item from a Bag, any proxies
 *	that have exposed UIs will update their UIs to indicate the change.
 *	
 *	<p>Unlike a Collection, a Bag also has a notion of the "top" element.
 *	(the one that you encounter first when you reach into the bag).
 *	The top element is typically modified by the user selecting
 *	an element in a UI created by a visual proxy. You can find out
 *	when that element changes by registering an ActionListener. You
 *	can also designate an existing bag element as the "top" element
 *	manually by calling {@link Bag#designate_top}.
 *	
 *	<p>As with standard collections, the Bag is not particularly thread 
 *	safe. It does take some trouble to make sure that the list of
 *	<code>ActionListener's</code> is handled in a thread-safe way, but
 *	that's it. If you want thread safety, you should wrap a <code>Bag</code>
 *	in a thread-safety wrapper of your own devising. You can use the
 *	methods of the <code>Collection</code> class for this purpose if
 *	you like, but only if you do not call any methods of <code>Bag</code>
 *	that are not defined in the <code>Collection</code> interface.
 *
 *	Though the <code>Bag</code> is not thread safe with respect to
 *	the outside world, it is thread safe with respect to Swing. That is,
 *	it's okay for the <code>Bag</code> to be modified, even if a proxy
 *	is exposing a UI.
 *	
 *	<p>Modifying the wrapped collection directly, rather than by calling a
 *	Bag method, is dangerous. If you need to modify an element, remove it,
 *	change it, then put it back.
 *	
 **/

public class Bag implements User_interface, Collection
{
	// A bug in the compiler (version 1.2.2 and all prior versions)
	// erroneously prints the error message "xxx may not have been
	// initialized," when the following are made final.  This bug
	// is somehow related to the presence of inner classes, but I'd
	// rather have the inner classes than the immutable fields.

	private /*final*/ Collection contents;
	private /*final*/ Comparator sort_strategy;
	private /*final*/ String	 name;
	private /*final*/ Vector	 proxies = new Vector();
	private			  Object	 top	 = null;

	/** Create a bag that uses the indicated Collection to represent
	 *  the indicated attribute.
	 */

	public Bag( Collection contents, String name  )
	{	this.contents		= contents;
		this.name		 	= name;
		this.sort_strategy	= null;
	}

	/** Create a <code>Bag</code> that wraps the <code>Collection</code>
	 *  passed in as an argument. When the items in the collection are
	 *	displayed, they are first extracted to an array, which is then
	 *	sorted using the supplied <code>Comparator</code>. (There's
	 *	no point in doing this if the underlying Collection is
	 *	already sorted, but it's handy for <code>HashSet</code> objects
	 *	and <code>LinkedList</code>.)
	 */

	public Bag( Collection contents, String name, Comparator sort_strategy )
	{	this.contents		= contents;
		this.name		    = name;
		this.sort_strategy	= sort_strategy;
	}

	/*==================================================================*/
	/* 				Pass-through (to Collection) methods.				*/
	/*==================================================================*/

    public boolean add	 (Object o)		  {	boolean result = contents.add(o);
    										changed(1);
    										return result;
    									  }
    public boolean addAll(Collection c)	  {	boolean result = contents.addAll(c);	
    										changed(1);
    										return result;
    									  }
    public boolean remove(Object o)	  	  {	boolean result = contents.remove(o);	
    										changed(-1);
    										return result;
    									  }
    public void    clear()	  			  {	contents.clear();	
    										changed(-1);
    									  }
    public boolean removeAll(Collection c){	boolean result = contents.removeAll(c);	
    										changed(-1);
    										return result;
    									  }
    public boolean retainAll(Collection c){	boolean result = contents.retainAll(c);	
    										changed(-1);
    										return result;
    									  }

    public int		size		()				{return contents.size();		}
    public boolean	isEmpty		()				{return contents.isEmpty();		}
    public boolean	contains	(Object o)		{return contents.contains(o);	}
    public Iterator iterator	()				{return contents.iterator();	}
    public Object[] toArray		()				{return contents.toArray();		}
    public Object[] toArray		(Object a[])	{return contents.toArray(a);	}
    public boolean	containsAll	(Collection c)	{return contents.containsAll(c);}
    public boolean	equals		(Object o)		{return contents.equals(o);		}
    public int		hashCode	()				{return contents.hashCode();	}

	/*==================================================================*/
	/* 					"Top" item management 							*/
	/*==================================================================*/

	/** Return the current "top" element.
	 */

	public Object top()
	{	return top;
	}

	private ActionListener subscription_list = null;

	/**	Add a listener that's notified when the designated "top" element
	 *	changes, either through a call to {@link #designate_top}
	 *	or by a user-mandated change made via a visual proxy.
	 */

	public final void addActionListener( ActionListener subscriber )
	{	subscription_list =
			AWTEventMulticaster.add( subscription_list, subscriber );
	}

	/** Remove a listener added by {@link #addActionListener}.
	 */

	public final void removeActionListener( ActionListener subscriber )
	{	subscription_list =
			AWTEventMulticaster.remove( subscription_list, subscriber );
	}

	/** Designate a new "top" element and notify any listeners that
	 *	the top item has changed.. Typically, this will be done
	 *	by the user picking an element in a proxy, but you can
	 *	do it manually by calling this method.
	 *	@param top the new "top" element. This element must be
	 *			an element of the encapsulated Collection, either added
	 *			directly or added via the <code>Bag</code> version
	 *			of {@link add()}.
	 *	@throws IllegalArgumentException if the argument doesn't
	 *			identify an item already in the bag.
	 */

	public void designate_top(final Object top)
	{
		if( !contents.contains( top ) )
			throw new IllegalArgumentException(
						"Designated top item not in Bag (" + top + ")" );
		this.top = top;

		if( subscription_list != null )
			subscription_list.actionPerformed
			(	new ActionEvent(this, ActionEvent.ACTION_PERFORMED,
														top.toString())
			);

		Object[] copy;
		synchronized(proxies){ copy = proxies.toArray(); }

		for( int i = 0; i < copy.length; ++i )	//#proxy_update
		{	final Proxy proxy = (Proxy)copy[i];
			SwingUtilities.invokeLater
			(	new Runnable()
				{	public void run()
					{	proxy.new_selection(top);
						proxy.repaint();
					}
				}
			);
		}
	}

	/*==================================================================*/
	/* 						  Visual proxy								*/
	/*==================================================================*/

	private void changed( final int direction )
	{
		Object[] copy;
		synchronized(proxies){ copy = proxies.toArray(); }

		for( int i = 0; i < copy.length; ++i )
		{	final Proxy proxy = (Proxy)copy[i];
			SwingUtilities.invokeLater
			(	new Runnable()
				{	public void run()
					{	proxy.changed( direction );
						proxy.repaint();
					}
				}
			);
		}
	}

	/** Manufacture a "visual proxy" for the current <code>Bag</code>,
	 *	suitable for inclusion in a "form". Currently, only one
	 *	proxy type (<code>"chooser"<code>) is supported.
	 *  The <code>Bag</code>
	 *	displays itself as a JList, JComboBox, or JButton, depending
	 *	on the amount of screen real estate available to it. In the
	 *	case of a button, the button label is the <code>attribute</code>
	 *	argument. Note that you can add an Ancestor listener to the
	 *	returned proxy to find out when the window that contains
	 *	the proxy shuts down. The most-recently selected item
	 *	can be fetched, at that point, by calling {@link #selected_item()}
	 *
	 *  @see Form
	 *  @see User_interface
	 */

	public JComponent visual_proxy( String attribute_name, boolean is_read_only )
	{	
		if( !is_read_only )
			return null;
	
		// Set things up so that the proxy will be in the "proxies"
		// list whenever it's visible. It's essential that it be removed
		// from the list when invisible; otherwise, the reference in
		// the list will keep the <code>Proxy</code> object from being garbage
		// collected after the containing window shuts down.

		final Proxy proxy = new Proxy( name );
		proxy.addAncestorListener
		(	new AncestorAdapter()
			{	public void ancestorRemoved(AncestorEvent event)	
				{	synchronized( proxies )
					{	proxies.remove( proxy );
					}
				}
				public void ancestorAdded(AncestorEvent event)
				{	synchronized( proxies )
					{	proxies.add( proxy );
					}
				}
			}
		);
		return proxy;
	}

	/** The actual visual-proxy class is a JPanel that contains either
	 *	a button, combo box, or list box, depending on its size. An
	 *	instance of this class is returned by the <code>visual_proxy()</code>
	 *	request.
	 */

	public class Proxy extends JPanel 
	{
		private int  	  	selected_size = 10;
		private String	  	attribute_name;
		private JComboBox 	drop_down;
		private JList	  	list;
		private JButton	  	button;
		private JComponent	current_ui;

		private Model 	 state  = new Model();
		private Renderer artist	= new Renderer();

		/** A private constructor, creates a proxy for the current
		 *	Bag. Since this constructor is private, you may not issue
		 *	a <code>new Bag("xxx")</code> request. Get a proxy by
		 *	calling the Bag's {@link Bag#visual_proxy} method.
		 *
		 *	@see Bag#visual_proxy.
		 */

		private Proxy( final String attribute_name )
		{
			drop_down	= new JComboBox	 ( state );
			list		= new JList		 ( state );
			button		= new JButton  	 ( attribute_name );

			drop_down.setRenderer(artist);
			drop_down.setSelectedIndex( 0 );

			// Set up the list. It turns out that the model, though
			// sufficient for controlling the appearance of the drop-
			// down, is not sufficient to control the list, so set
			// up a listener that sets modifies the state of the
			// current Bag in response to a selection.

			list.setCellRenderer (artist);
			list.addListSelectionListener
			(	new ListSelectionListener()
				{	public void valueChanged( ListSelectionEvent event )
					{	if( !event.getValueIsAdjusting() )
							designate_top(list.getSelectedValue());
					}
				}
			);

			// In the case of the button, set up a listener to throw up
			// the selection box (a small frame containing the JList)
			// when the button is pressed.

			button.addActionListener
			(	new ActionListener()
				{	public void actionPerformed( ActionEvent e )
					{	pop_up_selection_frame( attribute_name, button );
					}
				}
			);

			// Arrange for the UI to change it's appearance, if necessary,
			// when the size changes.

			this.addComponentListener	//#installcomponentlistener
			(	new ComponentAdapter()
				{	public void componentResized(ComponentEvent e)
					{	install_ui();
					}
				}
			);

			this.attribute_name = attribute_name;
			setLayout( new BorderLayout() );
		}

		/**	Examines the size of the <code>Proxy</code> object and
		 *	installs the button, drop-down, or list as appropriate.	
		 *	does nothing if the correct widget is already displayed.
		 */

		private final void install_ui()
		{
			Rectangle bounds 		 = getBounds();
			Dimension drop_down_size = drop_down.getPreferredSize();

			boolean use_list   = bounds.height > (drop_down_size.height *3 )
							  && bounds.width  > (drop_down_size.width  +10);
			boolean use_button = bounds.width  < drop_down_size.width 
							  || bounds.height < drop_down_size.height;
			boolean use_drop   = !use_list && !use_button;

			if( use_list 	&& current_ui == list		)	return;
			if( use_button	&& current_ui == button		)	return;
			if( use_drop	&& current_ui == drop_down	)	return;

			if( current_ui != null )
				current_ui.setVisible( false );
			this.removeAll();

			if( use_button )
				this.add( current_ui = button, BorderLayout.CENTER );
			else if( use_list )
				this.add( new JScrollPane(current_ui = list), BorderLayout.CENTER );
			else
			{	if( current_ui == null )
					drop_down.setSelectedIndex(0);
				this.add( current_ui = drop_down, BorderLayout.NORTH );
			}

			current_ui.setVisible( true );
			
			// Make the newly added component visible, note that
			// repaint(), invalidate(), and doLayout() do not work
			// for this purpose (probably a bug).

			this.setVisible( false );
			this.setVisible( true );
		}

		/** Called when the state of the underlying <code>Collection</code>
		 *	is changed by calling <code>add()</code>, <code>remove()</code>,
		 *	etc.
		 *	@param direction	&lt;o if the collection has gotten smaller,<br>
		 *						&gt;0 if it's gotten larger.<br>
		 *					    =0 if it hasn't changed size, but needs
		 *							to be refreshed.
		 */

		public void changed(int direction)
		{	int selected_index = list.getSelectedIndex();

			if( selected_index < 0 )
				selected_index = 0;

			if( selected_index >= contents.size() )
				selected_index = contents.size()-1;

			state.changed(direction);

			list.setSelectedIndex	 	(selected_index);
			list.ensureIndexIsVisible	(selected_index);

			drop_down.setSelectedIndex	(selected_index);
		}

		/** Called when the user picks an item from the current UI.
		 *	Makes sure that the drop-down and the list stay in synch
		 *	with each other.
		 */

		public void new_selection( Object selected )
		{	if( list.getSelectedValue()	 != selected )
			{	list.setSelectedValue	 ( selected, true );
				list.ensureIndexIsVisible( list.getSelectedIndex() );
			}
			if( drop_down.getSelectedItem() != selected )
			{	drop_down.setSelectedItem( selected );
			}
		}

		/** Overrides the base-class method, so must be public. Do not
		 *	call this method. Installs the ui when the proxy is
		 *	displayed the first time.
		 */

		public void addNotify()
		{	super.addNotify();
			install_ui();
		}

		/*******************************************************************
		 * The "model" class, holds the state of the font-name combo box.
		 * The list of possible fonts is stored here, as is the currently-selected
		 * font.
		 */

		private final class Model	extends		AbstractListModel
						 		    implements	ComboBoxModel
		{
			public Object getSelectedItem(		   ){ return top();	    	  }	
			public void   setSelectedItem(Object o ){ designate_top(o); 	  }

			public int	  getSize		 (		   ){ return contents.size(); }
			public Object getElementAt   (int index)
			{	Object[] items = contents.toArray();
				if( sort_strategy != null )
				{	Arrays.sort( items, sort_strategy );
				}
				return (0 <= index && index < items.length) ? items[index]: null;
			}

			/**	Call this method if the  collection changes in some way.
			 *	@param direction should be a negative number if the
			 *				collection got smaller. A positive number
			 *				if it got larger, 0 if it didn't change
			 *				size, but the UI needs redrawing anyway.
			 */

			public void changed( int direction )
			{	
				if( direction < 0 )
					fireIntervalAdded  (this, 0, contents.size()-1);
				else if( direction > 0 )
					fireIntervalRemoved(this, 0, contents.size()-1);
				else
					fireContentsChanged(this, 0, contents.size()-1);
			}
		}

	    /*******************************************************************
		 * The JList and JComboBox use the Renderer to draw its cells.
		 */

		private final class Renderer implements ListCellRenderer   
		{	
			private JPanel selected_wrapper = new JPanel();

			public Renderer()
			{	selected_wrapper.setLayout( new BorderLayout() );
				selected_wrapper.setBorder(
							BorderFactory.createLineBorder(Color.red) );
			}

			// Return a Component whose paint method is used to draw
			// the cell. Note that that's the only thing that the
			// Component is used for. The component itself it not put
			// into the cell.

			public Component getListCellRendererComponent(
										JList list, Object value, int index, 
										boolean isSelected, boolean cellHasFocus) 
			{
				JComponent item;

				if(value == null )
					value = "????";

				if( value instanceof User_interface ) 					//#proxy_choice
				{	item = ((User_interface)value).visual_proxy(null,true);
				}
				else
				{	item = new JLabel( value.toString() )
					{	public Dimension getPreferredSize()
						{	Dimension d = super.getPreferredSize();
							d.height += 8;
							d.width  += Math.min( d.width+8, 100 );
							return d;
						}
					};
				} 					//#proxy_choice_end

				if( isSelected )
				{	
					selected_wrapper.removeAll();
					selected_wrapper.add( item, BorderLayout.CENTER );
					item = selected_wrapper;
				}

				return item;
			}
		}

		/***************************************************************
		 * Create (and pop up) the "selection frame" that pop's up when
		 * the button UI is pressed. The (non-modal) frame contains a
		 * panel with an etched border, which in turn, holds the same
		 * JList object that appears when the Proxy has enough room
		 * to display it.
		 *
		 * The frame pops up over the button, rather than in the upper-left
		 * corner of the screen. The button is disabled as long as the
		 * selction frame is displayed. The selection frame is an
		 * JInternalFrame if this is an applet, otherwise it's a
		 * JFrame.
		 */

		private void pop_up_selection_frame( String attribute_name,
											final JButton button )
		{	
			button.setEnabled(false);
			Point location = button.getLocationOnScreen();
			location.x += 5;
			location.y += 5;

			JPanel interior = new JPanel();
			interior.setBorder
					(	BorderFactory.createTitledBorder
						(	new EtchedBorder(Color.white,Color.gray), 
							attribute_name
						)
					);
			interior.setLayout( new BorderLayout() );
			interior.add( new JScrollPane(list), BorderLayout.CENTER );
			list.setVisible( true );

			JApplet   applet= (JApplet) SwingUtilities.getAncestorOfClass
													(JApplet.class, button);
			JFrame    frame = (JFrame)  SwingUtilities.getAncestorOfClass
													(JFrame.class,  button);

			// Unfortunately, though JApplet and JFrame have many methods
			// in common, the two classes do not implement a common interface.
			// Similarly, the WindowListener and InternalFrameListener interfaces
			// are virtually identical, but again don't implement a common
			// interface. This lack of a comon base class is bad design,
			// but it's built into the Swing, and mandates that the following
			// if and else clauses must have seemingly identical code in places.
			// It didn't seem worth the effort to create an adapter to handle
			// this situation.

			if( applet != null )
			{
				final JInternalFrame selection_frame = new JInternalFrame(
														attribute_name, true );

				selection_frame.getContentPane().setLayout( new BorderLayout() );

				// A "closable" internal frame looks awful, and throws a mysterious
				// exception (under jdk 1.2.2.) when it's closed. Put a "close" button
				// at the bottom of the frame rather than using a "closable" window.

				JButton close = new JButton("Close");
				close.addActionListener
				(	new ActionListener()
					{	public void actionPerformed(ActionEvent e)
						{	button.setEnabled(true);
							selection_frame.dispose();
						}
					}
				);

				selection_frame.getContentPane().add( close,	BorderLayout.SOUTH );
				selection_frame.getContentPane().add( interior, BorderLayout.CENTER );

				// Put the internal frame into the "popup" layer of the JApplet's
				// root pane.

				applet.getRootPane().getLayeredPane().add(
									selection_frame , JLayeredPane.POPUP_LAYER);

				selection_frame.setLocation(location);
				selection_frame.pack();
				selection_frame.show();
			}
			else if( frame != null )
			{
				JFrame selection_frame = new JFrame( attribute_name );
				selection_frame.setContentPane( interior );
				selection_frame.addWindowListener
				(	new WindowAdapter()
					{	public void windowClosing(WindowEvent e)
						{	button.setEnabled(true);
						}
					}
				);

				selection_frame.setLocation(location);
				selection_frame.pack();
				selection_frame.show();
			}
			// else we can't pop up, do nothing.
		}
	}

	/*==================================================================*/
	/* 						  		TEST 								*/
	/*==================================================================*/

	/** This test class demonstrates how to use a Bag. It creates a
	 *	<code>Bag</code>, puts a few items in it, displays two UIs, then
	 *  allows you to add elements to the collection. You can resize the
	 *  windows to watch the display change from on appearance to another,
	 *	and when you type lines into the console window, the new lines
	 *	appear in both of the UI windows. Also notice that the proxies
	 *	stay in synch with each other with respect to selection as
	 *	well (If you don't want this last behavior, wrap the same
	 *	<code>Collection</code> object into two distinct <code>Bag</code>
	 *	objects.
	 *
	 *  <p>The class is a bit more complex than it needs to be in order to
	 *  make it easier to put the test panel into an applet. To test for
	 *  an application environment, run:
	 *	<pre>
	 *		java com.holub.ui.Bag\$Test
	 *	</pre>
	 *  To test an applet environment, run the test with:
	 *	<pre>
	 *		java com.holub.ui.Bag\$Test applet
	 *	</pre>
	 *  (Do not precede the $ with a \ if you're running under Windows.)
	 */

	public static class Test
	{
		public static JComponent create_ui( Collection aggregate, String title )
		{
			JInternalFrame frame = new JInternalFrame(title, true);

			// Get the visual proxy and shove it into the Frame

			User_interface displayable = (User_interface)aggregate;
			frame.getContentPane().add(
						displayable.visual_proxy("Attribute", true) );

			frame.pack();
			return frame;
		}

		public static JComponent test_panel()
		{	
			final Collection aggregate = new Bag( new LinkedList(), "Show List" );
			aggregate.add("A");
			aggregate.add("B");
			aggregate.add("C");
			aggregate.add("D");

			// You need to treat it as a Bag (as compared to a generic
			// Collection) to install an ActionListener, thus the cast.
			// It has to be final because it's referenced by the inner-
			// class object. Note that the listener will report all the
			// selections associated with displaying the initial UI.
			// There will be two such notifications (one for the list
			// and one for the drop-down) for each of the three proxies.
			// If you don't want this behavior, install the listener
			// after the visual proxy has been displayed.

			final JTextField status = new JTextField();

			final Bag the_bag = (Bag) aggregate;			//#add_listener
			the_bag.addActionListener
			(	new ActionListener()
				{	public void actionPerformed( ActionEvent e )
					{	status.setText( "Selected " + the_bag.top() );
					}
				}
			);

			JDesktopPane control_area = new JDesktopPane();
			control_area.setLayout( new FlowLayout() );
			control_area.add( create_ui(aggregate, "Proxy 1"), JLayeredPane.PALETTE_LAYER );
			control_area.add( create_ui(aggregate, "Proxy 2"), JLayeredPane.PALETTE_LAYER  );
			control_area.add( create_ui(aggregate, "Proxy 3"), JLayeredPane.PALETTE_LAYER  );
			control_area.setVisible(true);

			final JPanel input_area = new JPanel();
			final JTextField input = new JTextField();
			input.addActionListener
			(	new ActionListener()
				{	public void actionPerformed( ActionEvent e )
					{	aggregate.add( input.getText() );
						input.setText("");
					}
				}
			);
			input_area.setLayout( new GridBagLayout() );
			GridBagConstraints constraint = new GridBagConstraints();
			constraint.weightx 	 = 0;
			constraint.weighty 	 = 0;
			constraint.gridx   	 = 0;
			constraint.gridy   	 = 0;
			constraint.gridwidth = 1;
			constraint.gridheight= 1;
			constraint.fill      = GridBagConstraints.BOTH;
			input_area.add( new JLabel("Text to add:"), constraint );

			constraint.weightx 	 = 1;
			constraint.weighty 	 = 1;
			constraint.gridx   	 = 1;
			constraint.gridy   	 = 0;
			constraint.gridwidth = 1;
			constraint.gridheight= 1;
			constraint.fill      = GridBagConstraints.BOTH;
			input_area.add( input, constraint );

			input_area.setVisible(true);

			JPanel test_panel = new JPanel();
			test_panel.setLayout( new BorderLayout() );
			test_panel.add( status,			BorderLayout.NORTH  );
			test_panel.add( control_area,	BorderLayout.CENTER );
			test_panel.add( input_area,		BorderLayout.SOUTH  );
			test_panel.setVisible( true );

			return test_panel;
		}

		public static void main( String[] args ) throws Exception	//#main
		{ 	
			JFrame main_frame = new JFrame();

			if( args.length > 0 )
			{	System.out.println("Testing applet behavior");	
				JApplet applet = new JApplet();
				applet.getContentPane().add( test_panel() );
				main_frame.getContentPane().add( applet );
			}
			else
			{	System.out.println("Testing application behavior");		
				main_frame.getContentPane().add( test_panel() );
			}

			main_frame.addWindowListener
			(	new WindowAdapter()
				{	public void windowClosing( WindowEvent e )
					{	System.exit(0);
					}
				}
			);
			main_frame.pack();
			main_frame.show();
		}
	}
	//END_TEST
}
